/* mkfilter -- given n, compute recurrence relation
   to implement Butterworth, Bessel or Chebyshev filter of order n
   A.J. Fisher, University of York   <fisher@minster.york.ac.uk>
   September 1992 */

/* Main module */

#include <stdio.h>
#include "mkfilter.h"
#include "dspfilter.h"

#define opt_be 0x0001	/* -Be		Bessel cheracteristic	       */
#define opt_bu 0x0002	/* -Bu		Butterworth characteristic     */
#define opt_ch 0x0004	/* -Ch		Chebyshev characteristic       */

#define opt_lp 0x0008	/* -Lp		low-pass		       */
#define opt_hp 0x0010	/* -Hp		high-pass		       */
#define opt_bp 0x0020	/* -Bp		band-pass		       */

#define opt_a  0x0040	/* -a		alpha value		       */
#define opt_e  0x0100	/* -e		execute filter		       */
#define opt_l  0x0200	/* -l		just list filter parameters    */
#define opt_o  0x0400	/* -o		order of filter		       */
#define opt_p  0x0800	/* -p		specified poles only	       */
#define opt_w  0x1000	/* -w		don't pre-warp		       */

static int order, numpoles;
static double raw_alpha1, raw_alpha2;
static struct complex dc_gain, fc_gain, hf_gain;
static uint opts;	/* option flag bits */

static double warped_alpha1, warped_alpha2, chebrip;
static uint polemask;
static bool optsok;

static struct complex spoles[MAXPOLES];
static struct complex zpoles[MAXPOLES], zzeros[MAXPOLES];
static double xcoeffs[MAXPOLES+1], ycoeffs[MAXPOLES+1];

static struct complex bessel_poles[] =
  { /* table produced by /usr/fisher/bessel --	N.B. only one member of each C.Conj. pair is listed */
    { -1.000000e+00,  0.000000e+00 },	 { -8.660254e-01, -5.000000e-01 },    { -9.416000e-01,	0.000000e+00 },
    { -7.456404e-01, -7.113666e-01 },	 { -9.047588e-01, -2.709187e-01 },    { -6.572112e-01, -8.301614e-01 },
    { -9.264421e-01,  0.000000e+00 },	 { -8.515536e-01, -4.427175e-01 },    { -5.905759e-01, -9.072068e-01 },
    { -9.093907e-01, -1.856964e-01 },	 { -7.996542e-01, -5.621717e-01 },    { -5.385527e-01, -9.616877e-01 },
    { -9.194872e-01,  0.000000e+00 },	 { -8.800029e-01, -3.216653e-01 },    { -7.527355e-01, -6.504696e-01 },
    { -4.966917e-01, -1.002509e+00 },	 { -9.096832e-01, -1.412438e-01 },    { -8.473251e-01, -4.259018e-01 },
    { -7.111382e-01, -7.186517e-01 },	 { -4.621740e-01, -1.034389e+00 },    { -9.154958e-01,	0.000000e+00 },
    { -8.911217e-01, -2.526581e-01 },	 { -8.148021e-01, -5.085816e-01 },    { -6.743623e-01, -7.730546e-01 },
    { -4.331416e-01, -1.060074e+00 },	 { -9.091347e-01, -1.139583e-01 },    { -8.688460e-01, -3.430008e-01 },
    { -7.837694e-01, -5.759148e-01 },	 { -6.417514e-01, -8.175836e-01 },    { -4.083221e-01, -1.081275e+00 },
  };

static struct complex cmone = { -1.0, 0.0 };
static struct complex czero = {	 0.0, 0.0 };
static struct complex cone  = {	 1.0, 0.0 };
static struct complex ctwo  = {	 2.0, 0.0 };
static struct complex chalf = {	 0.5, 0.0 };

extern double pow(), tan(), atof(), fabs(), hypot(), atan2(), sqrt();
extern double sinh(), cosh(), asinh();

extern struct complex evaluate();	/* from complex.o */

uint decodeopts();
double getfarg();

extern struct complex csqrt(), cexp(), cconj(), cadd(), csub(), cmul(), cdiv();
#define cneg(z) csub(czero, z)


static void multin(struct complex w, struct complex coeffs[]);
static void calcfilter(t_filter *f, uint options, int ord, double rip, double a1, double a2);
static void checkoptions(void);
static void opterror(char *msg, int p1, int p2);
static void setdefaults(void);
static void compute_s(void); /* compute S-plane poles for prototype LP filter */
static void choosepole(struct complex z);
static void normalize(void);
static void compute_z(void); /* given S-plane poles, compute Z-plane poles */
static void expandpoly(void); /* given Z-plane poles & zeros, compute top & bot polynomials in Z, and then recurrence relation */
static void expand(struct complex pz[], struct complex coeffs[]);
static void printresults(char *argv[]);


double filter_filter(t_filter *f, double in)
{
    int i;
    double tmp;
    int np;

    np=f->numpoles;
    tmp=in * f->gain;
    for(i=0; i < np; i++)
    {
    	f->x[i]=f->x[i + 1];
    	f->y[i]=f->y[i + 1];
    }

    f->x[np]=tmp;
    
    for(i=0; i < np; i++)
    {
    	tmp+=((f->xcoeffs)[i] * (f->x)[i]) + ((f->ycoeffs)[i] * (f->y)[i]);
    }
    

    f->y[np]=tmp;
    
    return tmp;
}


void filter_butterworth(t_filter *f, int type, int ord, double a1, double a2)
{
    uint options;
    struct complex *g;
    
    options=opt_bu | opt_o | opt_a;
    switch(type)
    {
	case 0: options|=opt_lp; 
		g=&dc_gain;
		break;
	case 1: options|=opt_hp; 
		g=&hf_gain;
		break;
	case 2: options|=opt_bp; 
		break;
    }
    calcfilter(f, options, ord, 0.0, a1, a2);
    
    f->gain=1.0 / hypot(g->im, g->re);    	    
}

void filter_bessel(t_filter *f, int type, int ord, double a1, double a2)
{
    uint options;
    struct complex *g;
    
    options=opt_be | opt_o | opt_a;
    switch(type)
    {
	case 0: options|=opt_lp; 
		g=&dc_gain;
		break;
	case 1: options|=opt_hp; 
		g=&hf_gain;
		break;
	case 2: options|=opt_bp; 
		break;
    }


    calcfilter(f, options, ord, 0.0, a1, a2);
    f->gain=1.0 / hypot(g->im, g->re);    
}


void filter_chebyshev(t_filter *f, int type, int ord, double rip, double a1, double a2)
{
    uint options;
    struct complex *g;
    
    options=opt_ch | opt_o | opt_a;
    switch(type)
    {
	case 0: options|=opt_lp; 
		g=&dc_gain;
		break;
	case 1: options|=opt_hp; 
		g=&hf_gain;
		break;
	case 2: options|=opt_bp; 
		break;
    }


    calcfilter(f, options, ord, rip, a1, a2);
    f->gain=1.0 / hypot(g->im, g->re);    
}



void calcfilter(t_filter *f, uint options, int ord, double rip, double a1, double a2)
 { 
    int i;
    opts=options;
    order=ord;
    raw_alpha1=a1;
    raw_alpha2=a2;    
    chebrip=rip;

    checkoptions();
    setdefaults();
    compute_s();
    normalize();
    compute_z();
    expandpoly();
    
    for(i=0; i <= numpoles; i++)
    {
	(f->xcoeffs[i])=xcoeffs[i];
	(f->ycoeffs[i])=ycoeffs[i];
	f->x[i]=0.0;
	f->y[i]=0.0;
    }
    f->numpoles=numpoles;
  }


static void checkoptions()
  { uint m;
    optsok = true;
    m = opts & (opt_be | opt_bu | opt_ch);
    unless (m == opt_be || m == opt_bu || m == opt_ch)
	opterror("must specify exactly one of -Be, -Bu, -Ch", 0, 0);
    m = opts & (opt_lp | opt_hp | opt_bp);
    unless (m == opt_lp || m == opt_hp || m == opt_bp)
	opterror("must specify exactly one of -Lp, -Hp, -Bp", 0, 0);
    m = opts & (opt_l | opt_e);
    unless (m == opt_l || m == opt_e || m == 0)
	opterror("illegal combination of -l, -e", 0, 0);
    unless (opts & opt_o)
	opterror("must specify -o", 0, 0);
    unless (opts & opt_a)
	opterror("must specify -a", 0, 0);
    unless (order >= 1 && order <= MAXORDER)
	opterror("order must be in range 1 .. %d", MAXORDER, 0);
    if (opts & opt_p)
      { m = (1 << order) - 1; /* "order" bits set */
	if ((polemask & ~m) != 0)
	    opterror("order=%d, so args to -p must be in range 0 .. %d", order, order-1);
      }
    unless (optsok) exit(1);
  }

static void opterror(char *msg, int p1, int p2)
  { fprintf(stderr, "mkfilter: "); fprintf(stderr, msg, p1, p2); putc('\n', stderr);
    optsok = false;
  }

static void setdefaults()
  { unless (opts & opt_p) polemask = -1;
    unless (opts & opt_bp) raw_alpha2 = raw_alpha1;
  }

static void compute_s() /* compute S-plane poles for prototype LP filter */
  { numpoles = 0;
    if (opts & opt_be)
      { /* Bessel filter */
	int i;
	int p = (order*order)/4; /* ptr into table */
	if (order & 1) choosepole(bessel_poles[p++]);
	for (i=0; i < order/2; i++)
	  { choosepole(bessel_poles[p]);
	    choosepole(cconj(bessel_poles[p]));
	    p++;
	  }
      }
    if (opts & (opt_bu | opt_ch))
      { /* Butterworth filter */
	int i;
	for (i=0; i < 2*order; i++)
	  { struct complex s;
	    s.re = 0.0; s.im = (order & 1) ? (i*PI) / order : ((i+0.5)*PI) / order;
	    choosepole(cexp(s));
	  }
      }
    if (opts & opt_ch)
      { /* modify for Chebyshev (p. 136 DeFatta et al.) */
	double rip, eps, y; int i;
	if (chebrip >= 0.0)
	  { fprintf(stderr, "mkfilter: Chebyshev ripple is %g dB; must be .lt. 0.0\n", chebrip);
	    exit(1);
	  }
	rip = pow(10.0, -chebrip / 10.0);
	eps = sqrt(rip - 1.0);
	y = asinh(1.0 / eps) / (double) order;
	if (y <= 0.0)
	  { fprintf(stderr, "mkfilter: bug: Chebyshev y=%g; must be .gt. 0.0\n", y);
	    exit(1);
	  }
	for (i=0; i < numpoles; i++)
	  { spoles[i].re *= sinh(y);
	    spoles[i].im *= cosh(y);
	  }
      }
  }

static void choosepole(struct complex z)
  { if (z.re < 0.0)
      { if (polemask & 1) spoles[numpoles++] = z;
	polemask >>= 1;
      }
  }

static void normalize()
  { struct complex w1, w2; int i;
    /* for bilinear transform, perform pre-warp on alpha values */
    if (opts & opt_w)
      { warped_alpha1 = raw_alpha1;
	warped_alpha2 = raw_alpha2;
      }
    else
      { warped_alpha1 = tan(PI * raw_alpha1) / PI;
	warped_alpha2 = tan(PI * raw_alpha2) / PI;
      }
    w1.re = TWOPI * warped_alpha1; w1.im = 0.0;
    w2.re = TWOPI * warped_alpha2; w2.im = 0.0;
    /* transform prototype into appropriate filter type (lp/hp/bp) */
    switch (opts & (opt_lp + opt_hp + opt_bp))
      { case opt_lp:
	    for (i=0; i < numpoles; i++) spoles[i] = cmul(spoles[i], w1);
	    break;

	case opt_hp:
	    for (i=0; i < numpoles; i++) spoles[i] = cdiv(w1, spoles[i]);
	    /* also N zeros at (0,0) */
	    break;

	case opt_bp:
	  { struct complex w0, bw;
	    w0 = csqrt(cmul(w1, w2));
	    bw = csub(w2, w1);
	    for (i=0; i < numpoles; i++)
	      { struct complex hba, temp;
		hba = cmul(chalf, cmul(spoles[i], bw));
		temp = cdiv(w0, hba);
		temp = csqrt(csub(cone, cmul(temp, temp)));
		spoles[i] = cmul(hba, cadd(cone, temp));
		spoles[numpoles+i] = cmul(hba, csub(cone, temp));
	      }
	    /* also N zeros at (0,0) */
	    numpoles *= 2;
	    break;
	  }
      }
  }

static void compute_z() /* given S-plane poles, compute Z-plane poles */
  { int i;
    for (i=0; i < numpoles; i++)
      { /* use bilinear transform */
	struct complex top, bot;
	top = cadd(ctwo, spoles[i]);
	bot = csub(ctwo, spoles[i]);
	zpoles[i] = cdiv(top, bot);
	switch (opts & (opt_lp + opt_hp + opt_bp))
	  { case opt_lp:    zzeros[i] = cmone; break;
	    case opt_hp:    zzeros[i] = cone; break;
	    case opt_bp:    zzeros[i] = (i & 1) ? cone : cmone; break;
	  }
      }
  }

static void expandpoly() /* given Z-plane poles & zeros, compute top & bot polynomials in Z, and then recurrence relation */
  { struct complex topcoeffs[MAXPOLES+1], botcoeffs[MAXPOLES+1];
    struct complex st, zfc; int i;
    expand(zzeros, topcoeffs);
    expand(zpoles, botcoeffs);
    dc_gain = evaluate(topcoeffs, botcoeffs, numpoles, cone);
    st.re = 0.0; st.im = TWOPI * 0.5 * (raw_alpha1 + raw_alpha2); /* "jwT" for centre freq. */
    zfc = cexp(st);
    fc_gain = evaluate(topcoeffs, botcoeffs, numpoles, zfc);
    hf_gain = evaluate(topcoeffs, botcoeffs, numpoles, cmone);
    for (i=0; i <= numpoles; i++)
      { xcoeffs[i] = topcoeffs[i].re / botcoeffs[numpoles].re;
	ycoeffs[i] = -(botcoeffs[i].re / botcoeffs[numpoles].re);
      }
  }

static void expand(struct complex pz[], struct complex coeffs[])
  { /* compute product of poles or zeros as a polynomial of z */
    int i;
    coeffs[0] = cone;
    for (i=0; i < numpoles; i++) coeffs[i+1] = czero;
    for (i=0; i < numpoles; i++) multin(pz[i], coeffs);
    /* check computed coeffs of z^k are all real */
    for (i=0; i < numpoles+1; i++)
      { if (fabs(coeffs[i].im) > EPS)
	  { fprintf(stderr, "mkfilter: coeff of z^%d is not real; poles are not complex conjugates\n", i);
	    exit(1);
	  }
      }
  }

static void multin(struct complex w, struct complex coeffs[])
  { /* multiply factor (z-w) into coeffs */
    struct complex nw; int i;
    nw = cneg(w);
    for (i=numpoles; i >= 1; i--)
      coeffs[i] = cadd(cmul(nw, coeffs[i]), coeffs[i-1]);
    coeffs[0] = cmul(nw, coeffs[0]);
  }

static void printresults(char *argv[])
{ /* just list parameters */
	int i;
	printf("%d\n", numpoles);
	for (i=0; i <= numpoles; i++)
	  printf("%3g %14.10f\n", xcoeffs[i], ycoeffs[i]);
     
}

