#include <gtk/gtk.h>
#include "plugin.h"
#include <math.h>


#define MAX_DELAY 1000.0
#define MAX_SAMP_FREQ 45000
#define MAX_ECHO_DELAY 1.0

typedef struct echo_s
{
    float *buffer;
    int position;
    int size;
    int delay;
    float delay_time;
    float feedback;
    float smpf;
} echo_t;


struct plugin_data
{
	GtkWidget *config_window;
	echo_t *echo_left, *echo_right;
	float dryl, dryr;
	float cross;
	int sampling_frequency;
};


echo_t *echo_alloc(float delay_time, float fb, float smpfreq)
{
  echo_t *temp;

  temp=(echo_t *)malloc(sizeof(echo_t));
  temp->size=(int)(MAX_ECHO_DELAY * MAX_SAMP_FREQ);
  temp->buffer=(float *)malloc(sizeof(float) * temp->size);
  temp->position=0;
  temp->delay=(int)(delay_time/1000.0 * smpfreq);
  temp->feedback=fb;
  temp->smpf=smpfreq;
  temp->delay_time=delay_time;

  return temp;
}


void echo_free(echo_t *obj)
{
  free(obj->buffer);
  free(obj);
}


float echo_update(echo_t *obj, float data)
{
	float temp;
	int delayedsmp;

	delayedsmp=obj->position - obj->delay;
	if(delayedsmp < 0) delayedsmp+=obj->size;
	obj->buffer[obj->position++]=obj->buffer[delayedsmp] * obj->feedback + data;
	if(obj->position >= obj->size) obj->position=0;

	return obj->buffer[delayedsmp];
}

void echo_recalc(echo_t *obj, float delay_time, float fb, float smpf)
{
  obj->delay=(int)(delay_time/1000.0 * smpf);
  obj->delay_time=delay_time;
  obj->smpf=smpf;
  obj->feedback=fb;
}



static void cleanup(void *data)
{
	struct plugin_data *temp;

	temp=data;
	if(temp->config_window != NULL)
		gtk_widget_destroy(temp->config_window);
	echo_free(temp->echo_left);
	echo_free(temp->echo_right);	
	free(temp);
}

static void *init(void)
{
	struct plugin_data *temp;

	temp=(struct plugin_data *)malloc(sizeof(struct plugin_data));
	temp->echo_left=echo_alloc(MAX_DELAY / 2.0, 0.5, MAX_SAMP_FREQ);
	temp->echo_right=echo_alloc(MAX_DELAY / 2.0, 0.5, MAX_SAMP_FREQ);	
	temp->dryr=0.5;
	temp->dryl=0.5;	
	temp->cross=0.0;
	temp->sampling_frequency=MAX_SAMP_FREQ;
	temp->config_window=NULL;
	return temp;
}



static void delay_changed(GtkWidget *adj, echo_t *data)
{
	echo_recalc(data, (GTK_ADJUSTMENT(adj)->value), data->feedback, data->smpf);
}

static void dryleft_changed(GtkWidget *adj, struct plugin_data *data)
{
	data->dryl=GTK_ADJUSTMENT(adj)->value/100.0;
}

static void cross_changed(GtkWidget *adj, struct plugin_data *data)
{
	data->cross=GTK_ADJUSTMENT(adj)->value/100.0;
}


static void dryright_changed(GtkWidget *adj, struct plugin_data *data)
{
	data->dryr=GTK_ADJUSTMENT(adj)->value/100.0;
}



static void feedback_changed(GtkWidget *adj, echo_t *data)
{
	echo_recalc(data, data->delay_time, (GTK_ADJUSTMENT(adj)->value)/100.0, data->smpf);	
}


static void config_window_destroy(GtkWidget *w, gpointer data)
{
	struct plugin_data *temp;

	temp=data;

	if(temp->config_window != NULL)
		gtk_widget_destroy(temp->config_window);
	
	temp->config_window=NULL;
}


static void configure(void *data)
{
	struct plugin_data *temp;
	GtkWidget *vbox;
	GtkWidget *hbox;
        GtkWidget *label;	
	GtkObject *delayl_adj, *dryl_adj, *feedbackl_adj;
	GtkObject *delayr_adj, *dryr_adj, *feedbackr_adj;	
	GtkWidget *delay_bar, *dry_bar, *feedback_bar;
	GtkWidget *cross_bar;
	GtkObject *cross_adj;

	temp=data;
	
	if(temp->config_window != NULL) return;

	temp->config_window=gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(temp->config_window), "Digital Echo");
	gtk_widget_set_usize(temp->config_window, 300, 250);
	gtk_signal_connect(GTK_OBJECT(temp->config_window), "destroy", GTK_SIGNAL_FUNC(config_window_destroy), data);
	
	delayl_adj=gtk_adjustment_new((temp->echo_left)->delay_time, 5.0, 1000.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(delayl_adj), "value_changed", GTK_SIGNAL_FUNC(delay_changed), temp->echo_left);
	delayr_adj=gtk_adjustment_new((temp->echo_right)->delay_time, 5.0, 1000.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(delayr_adj), "value_changed", GTK_SIGNAL_FUNC(delay_changed), temp->echo_right);

	dryl_adj=gtk_adjustment_new(temp->dryl*100.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(dryl_adj), "value_changed", GTK_SIGNAL_FUNC(dryleft_changed), temp);
	dryr_adj=gtk_adjustment_new(temp->dryr*100.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(dryr_adj), "value_changed", GTK_SIGNAL_FUNC(dryright_changed), temp);

	feedbackl_adj=gtk_adjustment_new(temp->echo_left->feedback*100.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(feedbackl_adj), "value_changed", GTK_SIGNAL_FUNC(feedback_changed), temp->echo_left);
	feedbackr_adj=gtk_adjustment_new((temp->echo_right)->feedback*100.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(feedbackr_adj), "value_changed", GTK_SIGNAL_FUNC(feedback_changed), temp->echo_right);

	cross_adj=gtk_adjustment_new(temp->cross*100.0, 0.0, 100.0, 1.0, 1.0, 0.0);
	gtk_signal_connect(GTK_OBJECT(cross_adj), "value_changed", GTK_SIGNAL_FUNC(cross_changed), temp);


	hbox=gtk_hbox_new(FALSE, 0);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	label=gtk_label_new("Delay left(ms)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Delay right(ms)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Dry volume left(%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Dry volume right(%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);	
	label=gtk_label_new("Feedback left (%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);
	label=gtk_label_new("Feedback right (%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);

	label=gtk_label_new("Cross Feedback (%)");
	gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 10);
	gtk_widget_show(label);

	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	vbox=gtk_vbox_new(FALSE, 0);
	gtk_widget_show(vbox);
	delay_bar=gtk_hscale_new(GTK_ADJUSTMENT(delayl_adj));
	gtk_scale_set_digits(GTK_SCALE(delay_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(delay_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), delay_bar, TRUE, TRUE, 0);
	gtk_widget_show(delay_bar);	

	delay_bar=gtk_hscale_new(GTK_ADJUSTMENT(delayr_adj));
	gtk_scale_set_digits(GTK_SCALE(delay_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(delay_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), delay_bar, TRUE, TRUE, 0);
	gtk_widget_show(delay_bar);	


	dry_bar=gtk_hscale_new(GTK_ADJUSTMENT(dryl_adj));
	gtk_scale_set_digits(GTK_SCALE(dry_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(dry_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), dry_bar, TRUE, TRUE, 0);
	gtk_widget_show(dry_bar);	

	dry_bar=gtk_hscale_new(GTK_ADJUSTMENT(dryr_adj));
	gtk_scale_set_digits(GTK_SCALE(dry_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(dry_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), dry_bar, TRUE, TRUE, 0);
	gtk_widget_show(dry_bar);	

	
	feedback_bar=gtk_hscale_new(GTK_ADJUSTMENT(feedbackl_adj));
	gtk_scale_set_digits(GTK_SCALE(feedback_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(feedback_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), feedback_bar, TRUE, TRUE, 0);
	gtk_widget_show(feedback_bar);

	feedback_bar=gtk_hscale_new(GTK_ADJUSTMENT(feedbackr_adj));
	gtk_scale_set_digits(GTK_SCALE(feedback_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(feedback_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), feedback_bar, TRUE, TRUE, 0);
	gtk_widget_show(feedback_bar);

	cross_bar=gtk_hscale_new(GTK_ADJUSTMENT(cross_adj));
	gtk_scale_set_digits(GTK_SCALE(cross_bar), 0);
	gtk_scale_set_draw_value(GTK_SCALE(cross_bar), TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), cross_bar, TRUE, TRUE, 0);
	gtk_widget_show(cross_bar);	
	
	gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE ,0);


	gtk_widget_show(vbox);

	gtk_container_add(GTK_CONTAINER(temp->config_window), hbox);
	gtk_widget_show(temp->config_window);
}


static int process(void *data, short int *smpdata, int size, int sampling_frequency, int nchan)
{
	struct plugin_data *pdata;
	float sample_inl, sample_inr, sample_in;
	float sample_outl, sample_outr, sample_out;
	register int i;

	pdata=data;

	if(sampling_frequency != pdata->sampling_frequency)
	{
		pdata->sampling_frequency=sampling_frequency;
		echo_recalc(pdata->echo_left, pdata->echo_left->delay_time, pdata->echo_left->feedback, (float)sampling_frequency);
		echo_recalc(pdata->echo_right, pdata->echo_right->delay_time, pdata->echo_right->feedback, (float)sampling_frequency);		
	}
	
	if(nchan == 1)
	{
	    for(i=0; i < size; i++)
	    {
		sample_in=smpdata[i];
		
		sample_out=(1.0 - pdata->dryl) * echo_update(pdata->echo_left, sample_in);
		sample_out+=(1.0 - pdata->dryr) * echo_update(pdata->echo_right, sample_in);		
		sample_out+=sample_in * pdata->dryl + sample_in * pdata->dryr;
		
		if(sample_out > 32767.0) sample_out=32767.0;
		else if(sample_out < -32768.0) sample_out=-32768.0;
		
		smpdata[i]=(short int)sample_out;
	    }
	}
	else
	if(nchan == 2)
	    {
		for(i=0; i < size; i+=2)
		{
		    sample_inl=smpdata[i];
		    sample_inr=smpdata[i + 1];		    		    
		
		    sample_outl=(1.0 - pdata->dryl) * echo_update(pdata->echo_left, sample_inl);
		    sample_outr=(1.0 - pdata->dryr) * echo_update(pdata->echo_right, sample_inr);		
		    sample_outl+=sample_inl * pdata->dryl + sample_outr * pdata->cross;
		    sample_outr+=sample_inr * pdata->dryr + sample_outl * pdata->cross;
		
		    if(sample_outl > 32767.0) sample_outl=32767.0;
		    else if(sample_outl < -32768.0) sample_outl=-32768.0;

		    if(sample_outr > 32767.0) sample_outr=32767.0;
		    else if(sample_outr < -32768.0) sample_outr=-32768.0;
		
		    smpdata[i]=(short int)sample_outl;
		    smpdata[i + 1]=(short int)sample_outr;		    
		}
	    }
	
	return size;
}





Plugin plugin_info={"Digital Echo ",
					cleanup,
					init,
					configure,
					process};


Plugin *get_plugin_info(void)
{
	return &plugin_info;
}
